"""
This example demonstrates how to use a non-db model for a CRUD.

The model itself, doesn't really make sense here for a real-life app. But it
shows how you can organnize a CRUD around something that's not in the database
nor a managed Django model.

Objects are kept in-memory for this demonstration, but you could use any kind
of remote backend, you will only have to override the methods which are
overridden in this app.

So, basically, you need:

- a model with Meta.managed=False
- override get_queryset() and get_object() in the router,
- override save() and delete() in the model

Oh my, this was too easy, Django I <3 U WILL YOU MARRY ME !! Yes my friend this
is CRUDLFA+ and I can love and marry a software if that's how i feel like
hahahahahahahahahahah oh I'm sorry where you looking for boring software ? lol
"""


from crudlfap import crudlfap

from . import models


class SongRatingRouter(crudlfap.Router):
    def get_queryset(self, view):
        return self.model.objects.object_list

    def get_object(self, view):
        object_list = self.get_queryset(view)
        return object_list[int(view.kwargs['pk']) - 1]

SongRatingRouter(models.SongRating).register()



from crudlfap import crudlfap

from django.apps import apps
from django.conf import settings
from django.contrib.auth.models import Group

from . import views


User = apps.get_model(getattr(settings, 'AUTH_USER_MODEL', 'auth.User'))


def superuser(view):
    return view.request.user.is_superuser


crudlfap.Router(
    User,
    views=[
        crudlfap.DeleteObjectsView,
        crudlfap.DeleteView,
        crudlfap.UpdateView.clone(
            body_class='modal-fixed-footer',
            fields=[
                'username',
                'email',
                'first_name',
                'last_name',
                'groups',
                'is_superuser',
                'is_staff',
                'is_active',
            ]
        ),
        crudlfap.CreateView.clone(
            body_class='modal-fixed-footer',
            fields=['username', 'email', 'groups', 'is_staff', 'is_superuser'],
        ),
        views.PasswordView,
        views.BecomeUser,
        crudlfap.DetailView.clone(exclude=['password']),
        crudlfap.ListView.clone(
            search_fields=[
                'username',
                'email',
                'first_name',
                'last_name'
            ],
            table_fields=[
                'username',
                'email',
                'first_name',
                'last_name',
                'is_staff',
                'is_superuser'
            ],
            filter_fields=[
                'groups',
                'is_superuser',
                'is_staff'
            ],
        ),
    ],
    allowed=lambda view: view.request.user.is_superuser,
    urlfield='username',
    material_icon='person',
).register()

crudlfap.Router(
    Group,
    fields=['name', 'permissions'],
    material_icon='group',
).register()

crudlfap.site.views.append(views.Become.clone(model=User))



from django import http
from django.core.exceptions import ImproperlyConfigured
from django.db import models
from django.utils.translation import ugettext_lazy as _

from .model import ModelMixin


class ObjectMixin(ModelMixin):
    """Mixin for views using a Model instance."""

    menus = ['object', 'object_detail']
    menus_display = ['object', 'object_detail']
    object_permission_check = True
    template_name_field = None

    def get_context(self, **context):
        context['object'] = self.object
        return super().get_context(**context)

    def get_template_name_suffix(self):
        return '_{}'.format(self.urlname)

    def get_urlargs(self):
        """Return list with object's urlfield attribute."""
        return [getattr(self.object, self.urlfield)]

    @classmethod
    def to_url_args(cls, *args):
        """Return first arg's url_field attribute."""
        url_field = cls.get_url_field()
        return [getattr(args[0], url_field)]

    @classmethod
    def get_urlpath(cls):
        """Identify the object by slug or pk in the pattern."""
        return r'<{}>/{}'.format(cls.urlfield, cls.urlname)

    def get_title(self):
        return '{} "{}": {}'.format(
            self.model_verbose_name,
            self.object,
            _(self.view_label).capitalize(),
        ).capitalize()

    def get_menu_kwargs(self):
        return dict(object=self.object)

    def get_object(self, queryset=None):
        """
        Return the object the view is displaying.

        Require `self.queryset` and a `pk` or `slug` argument in the URLconf.
        Subclasses can override this to return any object.
        """
        if getattr(self, 'kwargs', False) is False:
            # This happens when the view has not been instanciated with an
            # object, neither from a URL which would allow getting the object
            # in the super() call below.
            raise Exception('Must instanciate the view with an object')

        # Use a custom queryset if provided; this is required for subclasses
        # like DateDetailView
        if queryset is None:
            queryset = self.get_queryset()

        queryset = queryset.filter(
            **{self.urlfield: self.kwargs.get(self.urlfield)}
        )

        try:
            # Get the single item from the filtered queryset
            obj = queryset.get()
        except queryset.model.DoesNotExist:
            raise http.Http404(
                _("No %(verbose_name)s found matching the query") %
                {'verbose_name': queryset.model._meta.verbose_name}
            )

        return obj

    def object_get(self):
        """Return the object, uses get_object() if necessary."""
        cached = getattr(self, '_object', None)
        if not cached:
            self._object = self.get_object()
        return self._object

    def object_set(self, value):
        """Set self.object attribute."""
        self._object = value

    object = property(object_get, object_set)

    def get_template_names(self):
        """
        Return a list of template names to be used for the request.

        May not be called if render_to_response() is overridden. Return the
        following list:

        * the value of ``template_name`` on the view (if provided)
        * the contents of the ``template_name_field`` field on the
          object instance that the view is operating upon (if available)
        * ``<app_label>/<model_name><template_name_suffix>.html``
        """
        try:
            names = super().get_template_names()
        except (ImproperlyConfigured, AttributeError):
            # If template_name isn't specified, it's not a problem --
            # we just start with an empty list.
            names = []

            # If self.template_name_field is set, grab the value of the field
            # of that name from the object; this is the most specific template
            # name, if given.
            if self.object and self.template_name_field:
                name = getattr(self.object, self.template_name_field, None)
                if name:
                    names.insert(0, name)

            # The least-specific option is the default
            # <app>/<model>_detail.html; only use this if the object in
            # question is a model.
            if isinstance(self.object, models.Model):
                object_meta = self.object._meta
                names.append("%s/%s%s.html" % (
                    object_meta.app_label,
                    object_meta.model_name,
                    self.template_name_suffix
                ))
            elif self.model and issubclass(self.model, models.Model):
                names.append("%s/%s%s.html" % (
                    self.model._meta.app_label,
                    self.model._meta.model_name,
                    self.template_name_suffix
                ))

            # If we still haven't managed to find any template names, we should
            # re-raise the ImproperlyConfigured to alert the user.
            if not names:
                raise

        return names



"""CRUD :eMixins that can be used in Actions or Views."""
import copy

from django import forms
from django.contrib.admin.models import ADDITION, CHANGE, DELETION, LogEntry
from django.contrib.contenttypes.models import ContentType


class CreateMixin:
    style = 'success'
    material_icon = 'add'
    default_template_name = 'crudlfap/create.html'
    controller = 'modal'
    action = 'click->modal#open'
    color = 'green'
    object_permission_check = False
    log_action_flag = ADDITION
    menus = ['main', 'model']
    short_permission_code = 'add'

    def form_valid(self):
        self.object = self.form.save()
        return super().form_valid()


class DeleteMixin:
    style = 'danger'
    fa_icon = 'trash'
    material_icon = 'delete'
    success_url_next = True
    color = 'red'
    log_action_flag = DELETION
    controller = 'modal'
    action = 'click->modal#open'
    form_class = forms.Form
    short_permission_code = 'delete'

    def form_valid(self):
        if hasattr(self, 'object_list'):
            self.result = copy.copy(self.object_list).delete()
        else:
            self.result = self.object.delete()
        return super().form_valid()

    def get_success_url(self):
        return self.router['list'].reverse()


class DetailMixin:
    fa_icon = 'search-plus'
    material_icon = 'search'
    default_template_name = 'crudlfap/detail.html'
    color = 'blue'
    menus_display = ['object', 'object_detail']

    @classmethod
    def get_urlpath(cls):
        """Identify the object by slug or pk in the pattern."""
        return r'<{}>'.format(cls.urlfield)

    def get_title(self):
        return str(self.object)

    def get_display_fields(self):
        self.display_fields = [
            {
                'field': self.model._meta.get_field(field),
                'value': getattr(self.object, field)
            }
            for field in (
                [f.name for f in self.model._meta.fields]
                if self.fields == '__all__'
                else self.fields
            ) if field not in self.exclude
        ]


class HistoryMixin:
    material_icon = 'history'
    template_name_suffix = '_history'
    default_template_name = 'crudlfap/history.html'
    controller = None
    action = None

    def get_object_list(self):
        ctype = ContentType.objects.get_for_model(self.model)
        return LogEntry.objects.filter(
            content_type=ctype,
            object_id=self.object.pk,
        )


class ListMixin:
    default_template_name = 'crudlfap/list.html'
    material_icon = 'list'
    body_class = 'full-width'
    menus = ['main', 'model']
    title_heading = None

    def get_urlpath(self):
        return ''

    def get_title_heading(self):
        return self.model._meta.verbose_name_plural.capitalize()


class UpdateMixin:
    material_icon = 'edit'
    default_template_name = 'crudlfap/update.html'
    controller = 'modal'
    action = 'click->modal#open'
    color = 'orange'
    locks = True
    log_action_flag = CHANGE
    short_permission_code = 'change'

    def get_form_fields(self):
        if hasattr(self, 'update_fields'):
            return self.update_fields
        if hasattr(self.router, 'update_fields'):
            return self.router.update_fields
        return super().get_form_fields()

    def form_valid(self):
        self.object = self.form.save()
        return super().form_valid()



import inspect


class FactoryMetaclass(type):
    def __getattr__(cls, attr):
        if attr.startswith('get_'):
            raise AttributeError('{} or {}'.format(attr[4:], attr))

        getter = getattr(cls, 'get_' + attr)

        if inspect.ismethod(getter):
            return getter()
        else:
            return getter(cls)

    def get_cls(cls):
        return cls


class Factory(metaclass=FactoryMetaclass):
    def __getattr__(self, attr):
        if attr.startswith('get_'):
            raise AttributeError('{} or {}()'.format(attr[4:], attr))

        getter = getattr(self, 'get_{}'.format(attr), None)

        if getter:
            methresult = getter()
            dictresult = self.__dict__.get(attr, None)
            if methresult is None and dictresult is not None:
                result = dictresult
            else:
                result = methresult
            return result

        # Try class methods
        return getattr(type(self), attr)

    @classmethod
    def clone(cls, **attributes):
        """Return a subclass with the given attributes.

        If a model is found, it will prefix the class name with the model.
        """
        name = cls.__name__
        model = attributes.get('model', getattr(cls, 'model', None))
        if model is None:
            model = getattr(cls, 'model', None)
        if model and model.__name__ not in cls.__name__:
            name = '{}{}'.format(model.__name__, cls.__name__)
        return type(name, (cls,), attributes)



from django import http
from django.conf import settings
from django.utils.module_loading import import_string

from .generic import TemplateView


class UrlsView(TemplateView):
    template_name = 'crudlfap/debug_urls.html'

    def printurl(self, parent, url):
        def url_repr(url):
            pattern = getattr(url, 'pattern', None)
            if pattern:
                regex = getattr(pattern, 'regex', None)
                if regex:
                    return regex.pattern.lstrip(
                        '^').rstrip('$').replace('\/', '/')
            return url_repr or ''
        ret = parent if parent else ''
        ret += url_repr(url)

        self.urlpatterns[ret] = url

        for pattern in getattr(url, 'url_patterns', []):
            self.printurl(ret, pattern)

    def get(self, request, *a, **k):
        urlpatterns = import_string(settings.ROOT_URLCONF).urlpatterns
        self.urlpatterns = {}
        for url in urlpatterns:
            self.printurl('/', url)

        if not request.user.is_superuser:
            return http.HttpResponseNotFound()
        return super(UrlsView, self).get(request, *a, **k)



"""
Crudlfa+ generic views and mixins.

Crudlfa+ takes views further than Django and are expected to:

- generate their URL definitions and reversions,
- check if a user has permission for an object,
- declare the names of the navigation menus they belong to.
"""
from django.conf import settings
from django.views import generic

from .. import mixins
from ..route import Route


if 'django.contrib.admin' in settings.INSTALLED_APPS:
    from django.contrib.admin.models import LogEntry
else:
    LogEntry = None


class View(mixins.LockMixin, mixins.MenuMixin, Route, generic.View):
    """Base view for CRUDLFA+."""


class TemplateView(mixins.TemplateMixin, View):
    """TemplateView for CRUDLFA+."""


class ModelView(mixins.ModelMixin, TemplateView):
    pass


class ObjectFormView(mixins.ObjectFormMixin, TemplateView):
    """Custom form view on an object."""


class ObjectView(mixins.ObjectMixin, TemplateView):
    pass


class ObjectsFormView(mixins.ObjectsFormMixin, TemplateView):
    pass


class ObjectsView(mixins.ObjectsMixin, TemplateView):
    pass


class FormView(mixins.FormMixin, TemplateView):
    """Base FormView class."""

    style = 'warning'
    default_template_name = 'crudlfap/form.html'


class ModelFormView(mixins.ModelFormMixin, FormView):
    pass


class CreateView(mixins.CreateMixin, ModelFormView):
    """View to create a model object."""


class DeleteView(mixins.DeleteMixin, ObjectFormView):
    """View to delete an object."""


class DeleteObjectsView(mixins.DeleteMixin, ObjectsFormView):
    """Delete selected objects."""


class DetailView(mixins.DetailMixin, ObjectView):
    """Templated model object detail view which takes a field option."""


class HistoryView(mixins.ObjectMixin, generic.DetailView):
    pass


class ListView(mixins.ListMixin, mixins.SearchMixin, mixins.FilterMixin,
               mixins.TableMixin, mixins.ObjectsMixin, TemplateView):

    def get_object_list(self):
        if self.filterset:
            self.object_list = self.filterset.qs
        else:
            self.object_list = self.queryset

        if self.search_form:
            self.object_list = self.search_form.get_queryset()
        return self.object_list

    def get_listactions(self):
        return self.router.get_menu('list_action', self.request)


class UpdateView(mixins.UpdateMixin, ObjectFormView):
    """Model update view."""



from django.core.exceptions import FieldDoesNotExist


def guess_urlfield(model):
    if not model:
        return None

    try:
        model._meta.get_field('slug')
    except FieldDoesNotExist:
        pass
    else:
        return 'slug'

    return 'pk'



from crudlfap import crudlfap

from .models import Post


class PostMixin:
    def get_exclude(self):
        if not self.request.user.is_staff:
            return ['owner']
        return super().get_exclude()


class PostCreateView(PostMixin, crudlfap.CreateView):
    def form_valid(self):
        self.form.instance.owner = self.request.user
        return super().form_valid()


class PostUpdateView(PostMixin, crudlfap.UpdateView):
    pass


class PostListView(crudlfap.ListView):
    def get_filter_fields(self):
        if self.request.user.is_staff:
            return ['owner']
        return []


class PostRouter(crudlfap.Router):
    fields = '__all__'
    icon = 'music'
    model = Post

    views = [
        crudlfap.DeleteObjectsView,
        crudlfap.DeleteView,
        PostUpdateView,
        PostCreateView,
        crudlfap.DetailView,
        PostListView.clone(
            search_fields=['name'],
        ),
    ]

    def allowed(self, view):
        """Example getting out of the django permission system."""
        user = view.request.user
        perms = view.required_permissions

        if perms == ['blog.add_post']:
            return user.is_authenticated
        elif perms == ['blog.change_post']:
            return view.object.editable(user)
        elif perms == ['blog.delete_post']:
            if hasattr(view, 'object'):
                return view.object.editable(user)

            # DeleteObjects relies on get_objects_for_user
            return user.is_authenticated

        return True

    def get_objects_for_user(self, user, perms):
        if perms in [['blog.change_post'], ['blog.delete_post']]:
            return self.model.objects.get_queryset().editable(user)
        elif perms in [['blog.list_post'], ['blog.detail_post']]:
            return self.model.objects.get_queryset().readable(user)
        else:
            return self.model.objects.get_queryset().none()

PostRouter().register()
